"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.blindDeserializePayload = exports.deserializePayload = exports.deserialize = exports.payloadDiscriminator = exports.serializePayload = exports.serialize = exports.registerPayloadTypes = exports.registerPayloadType = exports.createVAA = exports.composeLiteral = void 0;
const sdk_base_1 = require("@wormhole-foundation/sdk-base");
const layout_items_1 = require("./layout-items");
const utils_1 = require("./utils");
const composeLiteral = (protocol, payloadName) => (protocol ? `${protocol}:${payloadName}` : payloadName);
exports.composeLiteral = composeLiteral;
const decomposeLiteral = (payloadLiteral) => {
    const index = payloadLiteral.indexOf(":");
    return (index !== -1
        ? [payloadLiteral.slice(0, index), payloadLiteral.slice(index + 1)]
        : [null, payloadLiteral]);
};
const guardianSignatureLayout = [
    { name: "guardianIndex", binary: "uint", size: 1 },
    { name: "signature", ...layout_items_1.signatureItem },
];
const headerLayout = [
    { name: "version", binary: "uint", size: 1, custom: 1, omit: true },
    { name: "guardianSet", ...layout_items_1.guardianSetItem },
    {
        name: "signatures",
        binary: "array",
        lengthSize: 1,
        arrayItem: { binary: "object", layout: guardianSignatureLayout },
    },
];
//envelope + payload are getting hashed and signed
const envelopeLayout = [
    { name: "timestamp", binary: "uint", size: 4 },
    { name: "nonce", binary: "uint", size: 4 },
    { name: "emitterChain", ...(0, layout_items_1.chainItem)() },
    { name: "emitterAddress", ...layout_items_1.universalAddressItem },
    { name: "sequence", ...layout_items_1.sequenceItem },
    { name: "consistencyLevel", binary: "uint", size: 1 },
];
const baseLayout = [...headerLayout, ...envelopeLayout];
const payloadFactory = new Map();
function getPayloadLayout(layoutLiteral) {
    const layout = payloadFactory.get(layoutLiteral);
    if (!layout)
        throw new Error(`No layout registered for payload type ${layoutLiteral}`);
    return layout;
}
const payloadLiteralToPayloadItem = (payloadLiteral) => (payloadLiteral === "Uint8Array"
    ? { name: "payload", binary: "bytes" }
    : {
        name: "payload",
        binary: "object",
        layout: getPayloadLayout(payloadLiteral),
    });
function createVAA(payloadLiteral, vaaData) {
    const [protocolName, payloadName] = decomposeLiteral(payloadLiteral);
    const bodyLayout = [
        ...envelopeLayout,
        payloadLiteralToPayloadItem(payloadLiteral),
    ];
    const bodyWithFixed = (0, sdk_base_1.addFixedValues)(bodyLayout, 
    //not sure why the unknown cast here is required and why the type isn't inferred correctly
    vaaData);
    return {
        protocolName,
        payloadName,
        payloadLiteral,
        ...(0, sdk_base_1.addFixedValues)(headerLayout, vaaData),
        ...bodyWithFixed,
        hash: (0, utils_1.keccak256)((0, sdk_base_1.serializeLayout)(bodyLayout, bodyWithFixed)),
    };
}
exports.createVAA = createVAA;
function registerPayloadType(protocol, name, layout) {
    const payloadLiteral = (0, exports.composeLiteral)(protocol, name);
    if (payloadFactory.has(payloadLiteral))
        throw new Error(`Payload type ${payloadLiteral} already registered`);
    payloadFactory.set(payloadLiteral, layout);
}
exports.registerPayloadType = registerPayloadType;
function registerPayloadTypes(protocol, payloads) {
    for (const [name, layout] of payloads)
        registerPayloadType(protocol, name, layout);
}
exports.registerPayloadTypes = registerPayloadTypes;
function serialize(vaa) {
    const layout = [
        ...baseLayout,
        payloadLiteralToPayloadItem(vaa.payloadLiteral),
    ];
    return (0, sdk_base_1.serializeLayout)(layout, vaa);
}
exports.serialize = serialize;
function serializePayload(payloadLiteral, payload) {
    if (payloadLiteral === "Uint8Array")
        return payload;
    const layout = getPayloadLayout(payloadLiteral);
    return (0, sdk_base_1.serializeLayout)(layout, payload);
}
exports.serializePayload = serializePayload;
function payloadDiscriminator(payloadLiterals, allowAmbiguous) {
    const literals = (() => {
        if (Array.isArray(payloadLiterals[0]))
            return payloadLiterals.flatMap(([protocol, payloadNames]) => payloadNames.map((name) => (0, exports.composeLiteral)(protocol, name)));
        if (typeof payloadLiterals[1] === "string")
            return payloadLiterals;
        const [protocol, payloadNames] = payloadLiterals;
        return payloadNames.map((name) => (0, exports.composeLiteral)(protocol, name));
    })();
    const discriminator = (0, sdk_base_1.layoutDiscriminator)(literals.map((literal) => getPayloadLayout(literal)), !!allowAmbiguous);
    return ((data) => {
        if (typeof data === "string")
            data = sdk_base_1.encoding.hex.decode(data);
        const cands = discriminator(data);
        return Array.isArray(cands)
            ? cands.map((c) => literals[c])
            : cands !== null
                ? literals[cands]
                : null;
    });
}
exports.payloadDiscriminator = payloadDiscriminator;
function deserialize(payloadDet, data) {
    if (typeof data === "string")
        data = sdk_base_1.encoding.hex.decode(data);
    const [header, envelopeOffset] = (0, sdk_base_1.deserializeLayout)(headerLayout, data, 0, false);
    //ensure that guardian signature indicies are unique and in ascending order - see:
    //https://github.com/wormhole-foundation/wormhole/blob/8e0cf4c31f39b5ba06b0f6cdb6e690d3adf3d6a3/ethereum/contracts/Messages.sol#L121
    for (let i = 1; i < header.signatures.length; ++i)
        if (header.signatures[i].guardianIndex <=
            header.signatures[i - 1].guardianIndex)
            throw new Error("Guardian signatures must be in ascending order of guardian set index");
    const [envelope, payloadOffset] = (0, sdk_base_1.deserializeLayout)(envelopeLayout, data, envelopeOffset, false);
    const [payloadLiteral, payload] = typeof payloadDet === "string"
        ? [
            payloadDet,
            deserializePayload(payloadDet, data, payloadOffset),
        ]
        : deserializePayload(payloadDet, data, payloadOffset);
    const [protocolName, payloadName] = decomposeLiteral(payloadLiteral);
    const hash = (0, utils_1.keccak256)(data.slice(envelopeOffset));
    return {
        protocolName,
        payloadName,
        payloadLiteral,
        ...header,
        ...envelope,
        payload,
        hash,
    };
}
exports.deserialize = deserialize;
function deserializePayload(payloadDet, data, offset = 0) {
    //grouped together to have just a single cast on the return type
    return (() => {
        if (typeof data === "string")
            data = sdk_base_1.encoding.hex.decode(data);
        if (payloadDet === "Uint8Array")
            return data.slice(offset);
        if (typeof payloadDet === "string")
            return (0, sdk_base_1.deserializeLayout)(getPayloadLayout(payloadDet), data, offset);
        //kinda unfortunate that we have to slice here, future improvement would be passing an optional
        //  offset to the discriminator
        const candidate = payloadDet(data.slice(offset));
        if (candidate === null)
            throw new Error(`Encoded data does not match any of the given payload types - ${data}`);
        return [
            candidate,
            (0, sdk_base_1.deserializeLayout)(getPayloadLayout(candidate), data, offset),
        ];
    })();
}
exports.deserializePayload = deserializePayload;
exports.blindDeserializePayload = (() => {
    const rebuildDiscrimininator = () => {
        const layoutLiterals = Array.from(payloadFactory.keys());
        const layouts = layoutLiterals.map((l) => payloadFactory.get(l));
        return [layoutLiterals, (0, sdk_base_1.layoutDiscriminator)(layouts, true)];
    };
    let layoutLiterals = [];
    let discriminator = (_) => [];
    return (data) => {
        if (payloadFactory.size !== layoutLiterals.length)
            [layoutLiterals, discriminator] = rebuildDiscrimininator();
        if (typeof data === "string")
            data = sdk_base_1.encoding.hex.decode(data);
        const candidates = discriminator(data).map((c) => layoutLiterals[c]);
        return candidates.reduce((acc, literal) => {
            try {
                acc.push([
                    literal,
                    deserializePayload(literal, data),
                ]);
            }
            catch { }
            return acc;
        }, []);
    };
})();
//# sourceMappingURL=vaa.js.map